# Copyright (c) 2020, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
import numpy as np
import re
from ..core import platform_version
from .._services.model_repository import ModelRepository as modelRepo


# %%
class ScoreCode():

    @classmethod
    def writeScoreCode(cls, inputDF, targetDF, modelPrefix,
                       predictMethod, modelFileName,
                       metrics=['EM_EVENTPROBABILITY', 'EM_CLASSIFICATION'],
                       pyPath=Path.cwd(), threshPrediction=None,
                       otherVariable=False, model=None, isH2OModel=False, missingValues=False,
                       scoreCAS=True, isBinaryModel=False, binaryString=None):
        '''
        Writes a Python score code file based on training data used to generate the model
        pickle file. The Python file is included in the ZIP file that is imported or registered
        into the common model repository. The model can then be used by SAS applications,
        such as SAS Open Model Manager.

        The score code that is generated is designed to be a working template for any
        Python model, but is not guaranteed to work out of the box for scoring, publishing,
        or validating the model.

        Note that for categorical variables, the variable is split into the possible
        categorical values of the variable. Also, by default it does NOT include a catch-all
        [catVar]_Other variable to store any missing values or any values not found in the
        training data set. If you have missing values or values not included in your training
        data set, you must set the OtherVariable option to True.

        Both the inputDF and targetDF dataframes have the following stipulations:
        * Column names must be a valid Python variable name.
        * For categorical columns, the values must be a valid Python variable name.
        If either of these conditions is broken, an exception is raised.
        
        The following files are generated by this function:
        * '*Score.py'
            The Python score code file for the model.
        * 'dcmas_epscorecode.sas' (for SAS Viya 3.5 models)
            Python score code wrapped in DS2 and prepared for CAS scoring or publishing.
        * 'dmcas_packagescorecode.sas' (for SAS Viya 3.5 models)
            Python score code wrapped in DS2 and prepared for SAS Microanalyic Service scoring or publishing.

        Parameters
        ----------
        inputDF : DataFrame
            The `DataFrame` object contains the training data, and includes only the predictor
            columns. The writeScoreCode function currently supports int(64), float(64),
            and string data types for scoring.
        targetDF : DataFrame
            The `DataFrame` object contains the training data for the target variable.
        modelPrefix : string
            The variable for the model name that is used when naming model files.
            (For example: hmeqClassTree + [Score.py || .pickle]).
        predictMethod : string
            User-defined prediction method for score testing. This should be
            in a form such that the model and data input can be added using
            the format() command.
            For example: '{}.predict_proba({})'.
        modelFileName : string
            Name of the model file that contains the model.
        metrics : string list, optional
            The scoring metrics for the model. The default is a set of two
            metrics: EM_EVENTPROBABILITY and EM_CLASSIFICATION.
        pyPath : string, optional
            The local path of the score code file. The default is the current
            working directory.
        threshPrediction : float, optional
            The prediction threshold for probability metrics. For classification,
            below this threshold is a 0 and above is a 1.
        otherVariable : boolean, optional
            The option for having a categorical other value for catching missing
            values or values not found in the training data set. The default setting
            is False.
        model : str or dict
            The name or id of the model, or a dictionary representation of
            the model. The default value is None and is only necessary for models that
            will be hosted on SAS Viya 3.5.
        isH2OModel : boolean, optional
            Sets whether the model is an H2O.ai Python model. By default False.
        missingValues : boolean, optional
            Sets whether data used for scoring needs to go through imputation for
            missing values before passed to the model. By default False.
        scoreCAS : boolean, optional
    		Sets whether models registered to SAS Viya 3.5 should be able to be scored and
            validated through both CAS and SAS Micro Analytic Service. By default true. If
            set to false, then the model will only be able to be scored and validated through
            SAS Micro Analytic Service. By default True.
        isBinaryModel : boolean, optional
            Sets whether the H2O model provided is a binary model or a MOJO model. By default False.
        binaryString : string, optional
            Binary string representation of the model object. By default None.
        '''       
        # Check if binary string model
        if binaryString is not None:
            isBinaryString = True
        else:
            isBinaryString = False
        # Call REST API to check SAS Viya version
        isViya35 = (platform_version() == '3.5')

        # Initialize modelID to remove unbound variable warnings
        modelID = None

        # Helper method for uploading & migrating files
        # TODO: Integrate with register_model() or publish_model() task.
        def upload_and_copy_score_resources(model, files):
            for file in files:
                modelRepo.add_model_content(model, **file)
            return modelRepo.copy_python_resources(model)

        # For SAS Viya 3.5, either return an error or return the model UUID as a string
        if isViya35:
            if model == None:
                raise ValueError('The model UUID is required for score code written for' +
                                 ' SAS Model Manager on SAS Viya 3.5.')
            elif modelRepo.is_uuid(model):
                modelID = model
            elif isinstance(model, dict) and 'id' in model:
                modelID = model['id']
            else:
                model = modelRepo.get_model(model)
                modelID = model['id']

        # From the input dataframe columns, create a list of input variables, then check for viability
        inputVarList = list(inputDF.columns)
        for name in inputVarList:
            if not str(name).isidentifier():
                raise SyntaxError('Invalid column name in inputDF. Columns must be ' +
                                  'valid as Python variables.')
        newVarList = list(inputVarList)
        inputDtypesList = list(inputDF.dtypes)

        # Set the location for the Python score file to be written, then open the file
        zPath = Path(pyPath)
        pyPath = Path(pyPath) / (modelPrefix + 'Score.py')
        with open(pyPath, 'w') as cls.pyFile:

            # For H2O models, include the necessary packages
            if isH2OModel:
                cls.pyFile.write('''\
import h2o
import gzip, shutil, os''')
            # For binary string models, include the necessary packages
            if isBinaryString:
                cls.pyFile.write('''\
import codecs''')
            # Import math for imputation; pickle for serialized models; pandas for data management; numpy for computation    
            cls.pyFile.write('''\n
import math
import pickle
import pandas as pd
import numpy as np''')
            # In SAS Viya 4.0 and SAS Open Model Manager, a settings.py file is generated that points to the resource
            # location
            if not isViya35:
                cls.pyFile.write('''\n
import settings''')

            # Use a global variable for the model in order to load from memory only once
            cls.pyFile.write('''\n\n
_thisModelFit''')

            # For H2O models, include the server initialization, or h2o.connect() call to use an H2O server
            if isH2OModel:
                cls.pyFile.write('''\n
h2o.init()''')

            # For each case of SAS Viya version and H2O model or not, load the model file as variable _thisModelFit
            if isBinaryString:
                cls.pyFile.write('''\n
binaryString = """{binaryString}"""
_thisModelFit = pickle.loads(codecs.decode(binaryString.encode(), 'base64'))'''.format(binaryString=binaryString))
            elif (isViya35 and isH2OModel and not isBinaryModel):
                cls.pyFile.write('''\n
with gzip.open('/models/resources/viya/{modelID}/{modelFileName}', 'r') as fileIn, open('/models/resources/viya/{
modelID}/{modelZipFileName}', 'wb') as fileOut:
    shutil.copyfileobj(fileIn, fileOut)
os.chmod('/models/resources/viya/{modelID}/{modelZipFileName}', 0o777)
_thisModelFit = h2o.import_mojo('/models/resources/viya/{modelID}/{modelZipFileName}')'''.format(
                    modelID=modelID,
                    modelFileName=modelFileName,
                    modelZipFileName=modelFileName[:-4] + 'zip'
                ))
            elif isViya35 and not isH2OModel:
                cls.pyFile.write('''\n
with open('/models/resources/viya/{modelID}/{modelFileName}', 'rb') as _pFile:
    _thisModelFit = pickle.load(_pFile)'''.format(modelID=modelID, modelFileName=modelFileName))
            elif isViya35 and isBinaryModel:
                cls.pyFile.write('''\n
_thisModelFit = h2o.load_model('/models/resources/viya/{modelID}/{modelFileName}')'''.format(modelID=modelID, modelFileName=modelFileName))
            elif not isViya35 and not isH2OModel:
                cls.pyFile.write('''\n
with open(settings.pickle_path + '{modelFileName}', 'rb') as _pFile:
    _thisModelFit = pickle.load(_pFile)'''.format(modelFileName=modelFileName))
            elif not isViya35 and isBinaryModel:
                cls.pyFile.write('''\n
_thisModelFit = h2o.load_model(settings.pickle_path + '{}')'''.format(modelFileName=modelFileName))
            elif not isViya35 and isH2OModel and not isBinaryModel:
                cls.pyFile.write('''\n
with gzip.open(settings.pickle_path + '{modelFileName}', 'r') as fileIn, open(settings.pickle_path + '{
modelZipFileName}', 'wb') as fileOut:
    shutil.copyfileobj(fileIn, fileOut)
os.chmod(settings.pickle_path + '{modelZipFileName}', 0o777)
_thisModelFit = h2o.import_mojo(settings.pickle_path + '{modelZipFileName}')'''.format(modelFileName=modelFileName,
                                                                                       modelZipFileName=modelFileName[:-4] + 'zip'))
            # Create the score function with variables from the input dataframe provided and create the output variable line for SAS Model Manager
            cls.pyFile.write('''\n
def score{modelPrefix}({inputVarList}):
    "Output: {metrics}"'''.format(modelPrefix=modelPrefix,
                                  inputVarList=', '.join(inputVarList),
                                  metrics=', '.join(metrics)))
            # As a check for missing model variables, run a try/except block that reattempts to load the model in as a variable
            if binaryString is None:
                cls.pyFile.write('''\n
    try:
        global _thisModelFit
    except NameError:\n''')
                if isViya35 and not isH2OModel:
                    cls.pyFile.write('''
        with open('/models/resources/viya/{modelID}/{modelFileName}', 'rb') as _pFile:
            _thisModelFit = pickle.load(_pFile)'''.format(modelID=modelID, modelFileName=modelFileName))
                elif isViya35 and isH2OModel and not isBinaryModel:
                    cls.pyFile.write('''
        _thisModelFit = h2o.import_mojo('/models/resources/viya/{modelID}/{modelZipFileName}')
        '''.format(modelID=modelID, modelZipFileName=modelFileName[:-4] + 'zip'))
                elif isViya35 and isBinaryModel:
                    cls.pyFile.write('''
        _thisModelFit = h2o.load_model('/models/resources/viya/{modelID}/{modelFileName}')'''.format(modelID=modelID, modelFileName=modelFileName))
                elif not isViya35 and not isH2OModel:
                    cls.pyFile.write('''
        with open(settings.pickle_path + '{modelFileName}', 'rb') as _pFile:
            _thisModelFit = pickle.load(_pFile)'''.format(modelFileName=modelFileName))
                elif not isViya35 and isH2OModel:
                    cls.pyFile.write('''
        _thisModelFit = h2o.import_mojo(settings.pickle_path + '{}')'''.format(modelFileName[:-4] + 'zip'))
                elif not isViya35 and isBinaryModel:
                    cls.pyFile.write('''\n
        _thisModelFit = h2o.load_model(settings.pickle_path + '{}')'''.format(modelFileName=modelFileName))

            if missingValues:
                # For each input variable, impute for missing values based on variable dtype
                for i, dTypes in enumerate(inputDtypesList):
                    dTypes = dTypes.name
                    if 'int' in dTypes or 'float' in dTypes:
                        if cls.checkIfBinary(inputDF[inputVarList[i]]):
                            cls.pyFile.write('''\n
    try:
        if math.isnan({inputVar}):
            {inputVar} = {inputVarMode}
    except TypeError:
        {inputVar} = {inputVarMode}'''.format(inputVar=inputVarList[i],
                                              inputVarMode=float(list(inputDF[inputVarList[i]].mode())[0])))
                        else:
                            cls.pyFile.write('''\n
    try:
        if math.isnan({inputVar}):
            {inputVar} = {inputVarMean}
    except TypeError:
        {inputVar} = {inputVarMean}'''.format(inputVar=inputVarList[i],
                                              inputVarMean=float(inputDF[inputVarList[i]].mean(axis=0, skipna=True))))
                    elif 'str' in dTypes or 'object' in dTypes:
                        cls.pyFile.write('''\n
    try:
        categoryStr = {inputVar}.strip()
    except AttributeError:
        categoryStr = 'Other'\n'''.format(inputVar=inputVarList[i]))

                        tempVar = cls.splitStringColumn(inputDF[inputVarList[i]],
                                                        otherVariable)
                        newVarList.remove(inputVarList[i])
                        newVarList.extend(tempVar)
            # For non-H2O models, insert the model into the provided predictMethod call
            if not isH2OModel:
                predictMethod = predictMethod.format('_thisModelFit', 'inputArray')
                cls.pyFile.write('''\n
    try:
        inputArray = pd.DataFrame([[{newVars}]],
                                  columns=[{columns}],
                                  dtype=float)
        prediction = {predictMethod}
    except ValueError:
    # For models requiring or including an intercept value, a 'const' column is required
    # For example, many statsmodels models include an intercept value that must be included for the model prediction
        inputArray = pd.DataFrame([[1.0, {newVars}]],
                                columns=['const', {columns}],
                                dtype=float)
        prediction = {predictMethod}'''.format(newVars=', '.join(newVarList),
                                               columns=', '.join("'%s'" % x for x in newVarList),
                                               predictMethod=predictMethod)),
            elif isH2OModel:
                columnType = []
                for (var, dtype) in zip(newVarList, inputDtypesList):
                    if 'string' in dtype.name:
                        type = 'string'
                    else:
                        type = 'numeric'
                    columnType.append('\'' + var + '\'' + ':' + '\'' + type + '\'')
                cls.pyFile.write('''\n
    inputArray = pd.DataFrame([[{newVars}]],
                              columns=[{columns}],
                              dtype=float, index=[0])
    columnTypes = {{{columnTypes}}}
    h2oArray = h2o.H2OFrame(inputArray, column_types=columnTypes)
    prediction = _thisModelFit.predict(h2oArray)
    prediction = h2o.as_list(prediction, use_pandas=False)'''.format(newVars=', '.join(newVarList),
                                                                     columns=', '.join("'%s'" % x for x in newVarList),
                                                                     columnTypes=', '.join(columnType)
                                                                     ))
            if not isH2OModel:
                cls.pyFile.write('''\n
    try:
        {metric} = float(prediction)
    except TypeError:
    # If the model expects non-binary responses, a TypeError will be raised.
    # The except block shifts the prediction to accept a non-binary response.
        {metric} = float(prediction[:,1])'''.format(metric=metrics[0]))
                if threshPrediction is None:
                    threshPrediction = np.mean(targetDF)
                    cls.pyFile.write('''\n
    if ({metric0} >= {threshold}):
        {metric1} = '1'
    else:
        {metric1} = '0' '''.format(metric0=metrics[0],
                                   metric1=metrics[1],
                                   threshold=threshPrediction))
            elif isH2OModel:
                cls.pyFile.write('''\n
    {} = float(prediction[1][2])
    {} = prediction[1][0]'''.format(metrics[0], metrics[1]))
            
            metricsList = ', '.join(metrics)
            cls.pyFile.write('''\n
    return({})'''.format(metricsList))
            
            cls.pyFile.write('''\n''')
        
        # For SAS Viya 3.5, the model is first registered to SAS Model Manager, then the model UUID can be
        # added to the score code and reuploaded to the model file contents
        if isViya35:
            with open(pyPath, 'r') as pFile:
                files = [dict(name='{}Score.py'.format(modelPrefix), file=pFile, role='score')]
                upload_and_copy_score_resources(modelID, files)
            # After uploading the score code and migrating score resources, call the wrapper API to create
            # the Python score code wrapped in DS2
            modelRepo.convert_python_to_ds2(modelID)
            # Convert the DS2 wrapper code into two separate files: dmcas_epscorecode.sas and dmcas_packagescorecode.sas
            # The former scores or validates in CAS and the latter in SAS Microanalytic Service
            if scoreCAS:
                fileContents = modelRepo.get_model_contents(modelID)
                for item in fileContents:
                    if item.name == 'score.sas':
                        masCode = modelRepo.get('models/%s/contents/%s/content' % (item.modelId, item.id))
                with open(zPath / 'dmcas_packagescorecode.sas', 'w') as file:
                    print(masCode, file=file)
                casCode = cls.convertMAStoCAS(masCode, modelID)
                with open(zPath / 'dmcas_epscorecode.sas', 'w') as file:
                    print(casCode, file=file)
                for scoreCode in ['dmcas_packagescorecode.sas', 'dmcas_epscorecode.sas']:
                    with open(zPath / scoreCode, 'r') as file:
                        if scoreCode == 'dmcas_epscorecode.sas':
                            upload_and_copy_score_resources(modelID, [dict(name=scoreCode, file=file, role='score')])
                        else:
                            upload_and_copy_score_resources(modelID, [dict(name=scoreCode, file=file)])
                model = modelRepo.get_model(modelID)
                model['scoreCodeType'] = 'ds2MultiType'
                modelRepo.update_model(model)

    def splitStringColumn(cls, inputSeries, otherVariable):
        '''
        Splits a column of string values into a number of new variables equal
        to the number of unique values in the original column (excluding None
        values). It then writes to a file the statements that tokenize the newly
        defined variables.

        Here is an example: Given a series named strCol with values ['A', 'B', 'C',
        None, 'A', 'B', 'A', 'D'], designates the following new variables:
        strCol_A, strCol_B, strCol_D. It then writes the following to the file:
            strCol_A = np.where(val == 'A', 1.0, 0.0)
            strCol_B = np.where(val == 'B', 1.0, 0.0)
            strCol_D = np.where(val == 'D', 1.0, 0.0)

        Parameters
        ---------------
        inputSeries : string series
            Series with the string dtype.
        cls.pyFile : file (class variable)
            Open python file to write into.

        Returns
        ---------------
        newVarList : string list
            List of all new variable names split from unique values.
        '''

        uniqueValues = inputSeries.unique()
        uniqueValues = list(filter(None, uniqueValues))
        uniqueValues = [x for x in uniqueValues if str(x) != 'nan']
        newVarList = []
        for i, uniq in enumerate(uniqueValues):
            uniq = uniq.strip()
            if not uniq.isidentifier():
                raise SyntaxError('Invalid column value in inputDF. Values must be ' +
                                  'valid as Python variables (or easily space strippable).')
            newVarList.append('{}_{}'.format(inputSeries.name, uniq))
            cls.pyFile.write('''
    {0} = np.where(categoryStr == '{1}', 1.0, 0.0)'''.format(newVarList[i], uniq))

        if ('Other' not in uniqueValues) and otherVariable:
            newVarList.append('{}_Other'.format(inputSeries.name))
            cls.pyFile.write('''
    {}_Other = np.where(categoryStr == 'Other', 1.0, 0.0)'''.format(inputSeries.name))

        return newVarList

    def checkIfBinary(inputSeries):
        '''
        Checks a pandas series to determine whether the values are binary or nominal.

        Parameters
        ---------------
        inputSeries : float or int series
            A series with numeric values.

        Returns
        ---------------
        isBinary : boolean
            The returned value is True if the series values are binary, and False if the series values
            are nominal.
        '''

        isBinary = False
        binaryFloat = [float(1), float(0)]

        if inputSeries.value_counts().size == 2:
            if (binaryFloat[0] in inputSeries.astype('float') and
                    binaryFloat[1] in inputSeries.astype('float')):
                isBinary = False
            else:
                isBinary = True

        return isBinary

    def convertMAStoCAS(MASCode, modelId):
        '''Using the generated score.sas code from the Python wrapper API,
        convert the SAS Microanalytic Service based code to CAS compatible.

        Parameters
        ----------
        MASCode : str
            String representation of the packagescore.sas DS2 wrapper
        modelId : str or dict
            The name or id of the model, or a dictionary representation of
            the model

        Returns
        -------
        CASCode : str
            String representation of the epscorecode.sas DS2 wrapper code
        '''
        model = modelRepo.get_model(modelId)
        outputString = ''
        for outVar in model['outputVariables']:
            outputString = outputString + 'dcl '
            if outVar['type'] == 'string':
                outputString = outputString + 'varchar(100) '
            else:
                outputString = outputString + 'double '
            outputString = outputString + outVar['name'] + ';\n'
        start = MASCode.find('score(')
        finish = MASCode[start:].find(');')
        scoreVars = MASCode[start + 6:start + finish]
        inputString = ' '.join(
            [x for x in scoreVars.split(' ') if (x != 'double' and x != 'in_out' and x != 'varchar(100)')])
        endBlock = 'method run();\n    set SASEP.IN;\n    score({});\nend;\nenddata;'.format(inputString)
        replaceStrings = {'package pythonScore / overwrite=yes;': 'data sasep.out;',
                          'dcl int resultCode revision;': 'dcl double resultCode revision;\n' + outputString,
                          'endpackage;': endBlock}
        replaceStrings = dict((re.escape(k), v) for k, v in replaceStrings.items())
        pattern = re.compile('|'.join(replaceStrings.keys()))
        casCode = pattern.sub(lambda m: replaceStrings[re.escape(m.group(0))], MASCode)
        return casCode



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .service import Service


class Workflow(Service):
    """The Workflow API provides basic resources for list, prompt details,
    and running workflow processes.
    """
    _SERVICE_ROOT= '/workflow'

    def list_definitions(self,
                         include_enabled=True,
                         include_disabled=False):
        """List workflow definitions.

        Parameters
        ----------
        include_enabled : bool, optional
            Include enabled definitions in the results.  Defaults to True.
        include_disabled : bool, optional
            Include disabled definitions in the results.  Defaults to False.

        Returns
        -------
        list
            The list of definitions.

        """
        if include_enabled and include_disabled:
            url = '/definitions'
        elif include_enabled:
            url = '/enabledDefinitions'
        elif include_disabled:
            url = '/disabledDefinitions'
        else:
            return []

        # Header required to prevent 400 ERROR Bad Request
        return self.get(url, headers={"Accept-Language": "en-US"})

    def list_enabled_definitions(self):
        """List process definitions that are currently enabled.

        Returns
        -------
        list
            The list of definitions.

        """
        return self.list_definitions(include_enabled=True,
                                     include_disabled=False)

    def list_workflow_prompt(self, name):
        """List prompt Workflow Processes Definitions.

        Parameters
        ----------
        name : str
            Name or ID of an enabled workflow to retrieve inputs

        Returns
        -------
        list
            The list of prompts for specific workflow
    
        """
        
        ret = self._find_specific_workflow(name)
        if ret is None:
            raise ValueError("No Workflow enabled for %s name or id." % name)
        
        if 'prompts' in ret:
            return ret['prompts']
        else:
            # No prompt inputs on workflow
            return None
        
    def run_workflow_definition(self, name, input=None):
        """Runs specific Workflow Processes Definitions.

        Parameters
        ----------
        name : str
            Name or ID of an enabled workflow to execute
        input : dict, optional
            Input values for the workflow for initial workflow prompt

        Returns
        -------
        RestObj
            The executing workflow
            
        """

        workflow = self._find_specific_workflow(name)
        if workflow is None:
            raise ValueError("No Workflow enabled for %s name or id." % name)
        
        if input is None:
            return self.post('/processes?definitionId=' + workflow['id'],
	                     headers={'Content-Type': 'application/vnd.sas.workflow.variables+json'})
        if isinstance(input, dict):
            return self.post('/processes?definitionId=' + workflow['id'], json=input,
	                     headers={'Content-Type': 'application/vnd.sas.workflow.variables+json'})

    def _find_specific_workflow(self, name):
        # Internal helper method
        # Finds a workflow with the name (can be a name or id)
        # Returns a dict objects of the workflow
        listendef = self.list_enabled_definitions()
        for tmp in listendef:
            if tmp['name'] == name:
                return tmp
            elif tmp['id'] == name:
                return tmp
        
        # Did not find any enabled workflow with name/id
        return None





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Contains utilities for wrapping Python models in DS2 for publishing."""

from __future__ import print_function
import base64
import importlib
import pickle  # skipcq BAN-B301
import os
import re
import sys
from collections import OrderedDict


from .ds2 import DS2Thread, DS2Variable, DS2PyMASPackage
from .python import ds2_variables
from ..decorators import versionadded, versionchanged
from ..misc import random_string


@versionchanged(reason='Added `name` parameter.', version='1.5')
def build_wrapper_function(
    func, variables, array_input, name='wrapper', setup=None, return_msg=None
):
    """Wraps a function to ensure compatibility when called by PyMAS.

    PyMAS has strict expectations regarding the format of any function called
    directly by PyMAS.  Isolating the desired function inside a wrapping
    function provides a simple way to ensure that functions called by PyMAS
    are compliant.

    Parameters
    ----------
    func : function or str
        Function name or an instance of Function which will be wrapped
    variables : list of DS2Variable
    array_input : bool
        Whether `variables` should be combined into a single array before
        passing to `func`
    name : str, optional
        Name of the generated wrapper function.  Defaults to 'wrapper'.
    setup : iterable
        Python source code lines to be executed during package setup
    return_msg : bool, optional
        Deprecated.
        Whether error messages will be captured and returned as an additional
        output variable.  Defaults to True.

    Returns
    -------
    str
        the Python definition for the wrapper function.

    Notes
    -----
    The format for the `# Output: ` is very strict.  It must be exactly
    "# Output: <var>, <var>".  Any changes to spelling, capitalization,
    punctuation, or spacing will result in an error when the DS2 code is
    executed.

    """
    if return_msg is not None:
        raise DeprecationWarning(
            "The 'return_msg' parameter is ignored and "
            "will be removed completely in a future "
            "version"
        )

    # Always return an error message out of any Python method call.
    return_msg = True

    # need to know func & input/output vars for each func
    input_names = [v.name for v in variables if not v.out]
    output_names = [v.name for v in variables if v.out]
    args = input_names
    func = func.__name__ if callable(func) else func

    # HELPER: SAS to python char issue where SAS char have spaces and python string does not.
    # NOTE: we assume SAS char always need white space to be trimmed.  This seems to match python model built so far
    # Note: Using this string fix to help with None or blank input situation for numerics
    string_input = ('',)
    for v in variables:
        if not v.out:
            if v.type == 'char':
                string_input += ("        if {0}: {0} = {0}.strip()".format(v.name),)
            else:
                string_input += ("        if {0} is None: {0} = np.nan".format(v.name),)

    # Statement to execute the function w/ provided parameters
    if array_input:
        middle = string_input + (
            '        input_array = np.array([{}]).reshape((1, -1))'.format(
                ', '.join(args)
            ),
            '        columns = [{}]'.format(', '.join('"{0}"'.format(w) for w in args)),
            '        input_df = pd.DataFrame(data=input_array, columns=columns)',
            '        result = {}(input_df)'.format(func),
        )
    else:
        func_call = '{}({})'.format(func, ','.join(args))
        middle = ('        result = {}'.format(func_call),)

    if setup:
        header = (
            ('try:',)
            + tuple('    ' + line for line in setup)
            + (
                '    _compile_error = None',
                'except Exception as e:',
                '    _compile_error = e',
                '',
            )
        )
    else:
        header = ('',)

    # NOTE: 'Output:' section is required.  All return variables must be listed
    # separated by ', '
    definition = (
        header
        + (
            'def {name}({args}):'.format(name=name, args=', '.join(args)),
            '    "Output: {}"'.format(
                ', '.join(output_names + ['msg'])
                if return_msg
                else ', '.join(output_names)
            ),
            '    result = None',
            '    msg = None' if return_msg else '',
            '    try:',
            '        global _compile_error',
            '        if _compile_error is not None:',
            '            raise _compile_error',
            '        import numpy as np',
            '        import pandas as pd',
        )
        + middle
        + (
            '        result = tuple(result.ravel()) if hasattr(result, '
            '"ravel") else tuple(result)',
            '        if len(result) == 0:',
            '            result = tuple(None for i in range(%s))' % len(output_names),
            '        elif "numpy" in str(type(result[0])):',
            '            result = tuple(np.asscalar(i) for i in result)',
            '    except Exception as e:',
            '        from traceback import format_exc',
            '        msg = str(e) + format_exc()' if return_msg else '',
            '        if result is None:',
            '            result = tuple(None for i in range(%s))' % len(output_names),
            '    return result + (msg, )',
        )
    )

    return '\n'.join(definition)


@versionadded(version='1.5')
def wrap_predict_method(func, variables, **kwargs):
    """Create a PyMAS wrapper designed for Sci-kit's `.predict` methods.

    Parameters
    ----------
    func : function or str
        Function name or an instance of Function which will be wrapped.  Assumed
        to behave as `.predict()` methods.
    variables : list of DS2Variable
        Input and output variables for the function
    kwargs : any
        Will be passed to `build_wrapper_function`.

    Returns
    -------
    str
        the Python definition for the wrapper function.

    See Also
    --------
    :meth:`build_wrapper_function`

    """
    kwargs.setdefault('array_input', True)
    kwargs.setdefault('name', 'predict')

    return build_wrapper_function(func, variables, **kwargs)


@versionadded(version='1.5')
def wrap_predict_proba_method(func, variables, **kwargs):
    """Create a PyMAS wrapper designed for Sci-kit's `.predict_proba` methods.

    Parameters
    ----------
    func : function or str
        Function name or an instance of Function which will be wrapped.  Assumed
        to behave as `.predict_proba()` methods.
    variables : list of DS2Variable
        Input and output variables for the function
    kwargs : any
        Will be passed to `build_wrapper_function`.

    Returns
    -------
    str
        the Python definition for the wrapper function.

    See Also
    --------
    :meth:`build_wrapper_function`

    """
    kwargs.setdefault('array_input', True)
    kwargs.setdefault('name', 'predict_proba')

    wrapper = build_wrapper_function(func, variables, **kwargs)

    old_code = r'if result.size == 1:\s*result = np.asscalar\(result\)'
    new_code = 'assert result.shape[0] == 1\n'
    new_code += '        result = tuple(result[0].tolist())'

    return re.sub(old_code, new_code, wrapper)


@versionchanged('Return code and message are disabled by default.', version='1.5')
def from_inline(
    func, input_types=None, array_input=False, return_code=None, return_message=None
):
    """Creates a PyMAS wrapper to execute the inline python function.

    Parameters
    ----------
    func : function
        A Python function object to be used
    input_types : list of type, optional
        The expected type for each input value of `func`.  Can be ommitted if
        `func` includes type hints.
    array_input : bool
        Whether the function inputs should be treated as an array instead of
        individual parameters
    return_code : bool
        Deprecated.
        Whether the DS2-generated return code should be included
    return_message : bool
        Deprecated.
        Whether the DS2-generated return message should be included

    Returns
    -------
    PyMAS
        Generated DS2 code which can be executed in a SAS scoring environment

    """

    if return_code is not None or return_message is not None:
        raise DeprecationWarning(
            "The 'return_code' and 'return_message' "
            "parameters are ignored and will be "
            "removed completely in a future version"
        )

    obj = pickle.dumps(func)
    return from_pickle(obj, None, input_types, array_input, return_code, return_message)


@versionchanged('Return code and message are disabled by default.', version='1.5')
def from_python_file(
    file,
    func_name=None,
    input_types=None,
    array_input=False,
    return_code=None,
    return_message=None,
):
    """Creates a PyMAS wrapper to execute a function defined in an
    external .py file.

    Parameters
    ----------
    file : str
        The path to a python source code file
    func_name : str
        Name of the target function to call
    input_types : list of type, optional
        The expected type for each input value of the target function.
        Can be ommitted if target function includes type hints.
    array_input : bool
        Whether the function inputs should be treated as an array instead of
        individual parameters
    return_code : bool
        Deprecated
        Whether the DS2-generated return code should be included
    return_message : bool
        Deprecated
        Whether the DS2-generated return message should be included

    Returns
    -------
    PyMAS
        Generated DS2 code which can be executed in a SAS scoring environment

    """
    if return_code is not None or return_message is not None:
        raise DeprecationWarning(
            "The 'return_code' and 'return_message' "
            "parameters are ignored and will be "
            "removed completely in a future version"
        )

    if not str(file.lower().endswith('.py')):
        raise ValueError("File {} does not have a .py extension.".format(file))

    # Extract just the filename from the path
    directory, file_name = os.path.split(file)

    # Drop the file extension to get the module name
    module_name = os.path.splitext(file_name)[0]

    # Temporarily add the module location to Python's path
    sys.path.append(directory)
    module = importlib.import_module(module_name)
    sys.path.pop()
    target_func = getattr(module, func_name)

    if not callable(target_func):
        raise RuntimeError("Could not find a valid function named %s" % func_name)

    with open(file, 'r') as f:
        code = [line.strip('\n') for line in f.readlines()]

    return _build_pymas(target_func, None, input_types, array_input, code=code)


@versionchanged('Return code and message are disabled by default.', version='1.5')
def from_pickle(
    file,
    func_name=None,
    input_types=None,
    array_input=False,
    return_code=None,
    return_message=None,
):
    """Create a deployable DS2 package from a Python pickle file.

    Parameters
    ----------
    file : str or bytes or file_like
        Pickled object to use.  String is assumed to be a path to a picked
        file, file_like is assumed to be an open file handle to a pickle
        object, and bytes is assumed to be the raw pickled bytes.
    func_name : str
        Name of the target function to call
    input_types : DataFrame, type, list of type, or dict of str: type, optional
        The expected type for each input value of the target function.
        Can be omitted if target function includes type hints.  If a DataFrame
        is provided, the columns will be inspected to determine type information.
        If a single type is provided, all columns will be assumed to be that type,
        otherwise a list of column types or a dictionary of column_name: type
        may be provided.
    array_input : bool
        Whether the function inputs should be treated as an array instead of
        individual parameters
    return_code : bool
        Deprecated.
        Whether the DS2-generated return code should be included
    return_message : bool
        Deprecated.
        Whether the DS2-generated return message should be included

    Returns
    -------
    PyMAS
        Generated DS2 code which can be executed in a SAS scoring environment

    """
    if return_code is not None or return_message is not None:
        raise DeprecationWarning(
            "The 'return_code' and 'return_message' "
            "parameters are ignored and will be "
            "removed completely in a future version"
        )
    try:
        # In Python2 str could either be a path or the binary pickle data,
        # so check if its a valid filepath too.
        is_file_path = isinstance(file, str) and os.path.isfile(file)
    except TypeError:
        is_file_path = False

    # Path to a pickle file
    if is_file_path:
        with open(file, 'rb') as f:
            obj = pickle.load(f)  # skipcq BAN-B301

    # The actual pickled bytes
    elif isinstance(file, bytes):
        obj = pickle.loads(file)  # skipcq BAN-B301
    else:
        obj = pickle.load(file)  # skipcq BAN-B301

    # Encode the pickled data so we can inline it in the DS2 package
    pkl = base64.b64encode(pickle.dumps(obj))

    code = (
        'import pickle, base64',
        # Replace b' with " before embedding in DS2.
        'bytes = {}'.format(pkl).replace("'", '"'),
        'obj = pickle.loads(base64.b64decode(bytes))',
    )

    return _build_pymas(
        obj, func_name, input_types, array_input, func_prefix='obj.', code=code
    )


def _build_pymas(
    obj,
    func_name=None,
    input_types=None,
    array_input=False,
    func_prefix=None,
    return_code=None,
    return_message=None,
    code=None,
):
    """

    Parameters
    ----------
    obj
    func_name
    input_types
    array_input
    return_code
    return_message
    code

    Returns
    -------

    """
    code = code or []

    if not isinstance(func_name, list):
        func_name = [func_name]

    def parse_function(obj, func_name):
        # If the object passed was a function, no need to search for
        # target function
        if callable(obj) and (func_name is None or obj.__name__ == func_name):
            target_func = obj
        elif func_name is None:
            raise ValueError('Parameter `func_name` must be specified.')
        else:
            target_func = getattr(obj, func_name)

        if not callable(target_func):
            raise RuntimeError("Could not find a valid function named %s" % func_name)

        target_func_name = target_func.__name__

        if target_func_name.lower() == 'predict_proba':
            names = 'P_'
        else:
            names = None

        # Need to create DS2Variable instances to pass to PyMAS
        if hasattr(input_types, 'columns'):
            # Assuming input is a DataFrame representing model inputs.  Use to
            # get input variables
            vars = ds2_variables(input_types)

            # Run one observation through the model and use the result to
            # determine output variables
            output = target_func(input_types.head(1))
            output_vars = ds2_variables(output, output_vars=True, names=names)
            vars.extend(output_vars)
        elif isinstance(input_types, type):
            params = OrderedDict(
                [(k, input_types) for k in target_func.__code__.co_varnames]
            )
            vars = ds2_variables(params)
        elif isinstance(input_types, dict):
            vars = ds2_variables(input_types)
        else:
            # Inspect the Python method to determine arguments
            vars = ds2_variables(target_func)

        if not any(v for v in vars if v.out):
            vars.append(DS2Variable(name='result', type='float', out=True))

        return target_func_name, vars

    # Get variables for each function
    temp = [parse_function(obj, f) for f in func_name]
    target_func, variables = zip(*temp)

    # Convert from tuples to lists
    target_func = list(target_func)
    variables = list(variables)

    return PyMAS(
        target_func, variables, code, array_input=array_input, func_prefix=func_prefix
    )


class PyMAS:
    """

    Parameters
    ----------
    target_function : str
        The Python function to be executed.
    variables : list of DS2Variable
        The input/ouput variables be declared in the module.
    python_source : str
        Additional Python code to be executed during setup.
    return_code : bool
        Deprecated.
        Whether the DS2-generated return code should be included.
    return_msg : bool
        Deprecated.
        Whether the DS2-generated return message should be included.
    kwargs : any
        Passed to :func:`build_wrapper_function`.

    """

    def __init__(
        self,
        target_function,
        variables,
        python_source,
        return_code=None,
        return_msg=None,
        func_prefix=None,
        **kwargs
    ):

        if return_code is not None or return_msg is not None:
            raise DeprecationWarning(
                "The 'return_code' and 'return_msg' "
                "parameters are ignored and will be "
                "removed completely in a future version"
            )

        func_prefix = func_prefix or ''

        if not isinstance(target_function, list):
            target_function = [target_function]
            variables = [variables]

        return_code = False
        return_msg = False

        # Replicate parameter for each function to be exposed
        if isinstance(return_code, bool):
            return_code = [return_code] * len(target_function)

        # Replicate parameter for each function to be exposed
        if isinstance(return_msg, bool):
            return_msg = [return_msg] * len(target_function)

        self.wrapper = []
        wrapper_names = []
        for func, vars, code, msg in zip(
            target_function, variables, return_code, return_msg
        ):
            wrapper_names.append('_' + random_string(20))

            if func.lower() == 'predict':
                lines = wrap_predict_method(
                    func_prefix + func,
                    vars,
                    setup=python_source,
                    name=wrapper_names[-1],
                    **kwargs
                )
            elif func.lower() == 'predict_proba':
                lines = wrap_predict_proba_method(
                    func_prefix + func,
                    vars,
                    name=wrapper_names[-1],
                    setup=python_source,
                    **kwargs
                )
            else:
                lines = build_wrapper_function(
                    func_prefix + func,
                    vars,
                    name=wrapper_names[-1],
                    setup=python_source,
                    **kwargs
                )

            # Add DS2 variables for returning error codes/messages
            # NOTE: add these *after* wrapper function is generated to prevent
            # double-counting them.
            if code:
                vars.append(DS2Variable(name='rc', type='int32', out=True))
            if msg:
                vars.append(DS2Variable(name='msg', type='char', out=True))

            # Clear setup code once it's been added once.  No need to duplicate
            # if multiple functions are defined.
            python_source = None

            self.wrapper.extend(lines.split('\n'))

        self.variables = variables[0]
        # self.return_code = return_code[0]
        self.return_message = return_msg[0]

        self.package = DS2PyMASPackage(self.wrapper)

        for idx, func in enumerate(target_function):
            self.package.add_method(func, wrapper_names[idx], variables[idx])

    @versionchanged(version='1.4', reason="Added `dest='Python'` option")
    def score_code(self, input_table=None, output_table=None, columns=None, dest='MAS'):
        """Generate DS2 score code

        Parameters
        ----------
        input_table : str
            The name of the table to execute the function against
        output_table : str
            The name of the table where execution results will be written
        columns : list of str
            Names of the columns from `table` that will be passed to `func`
        dest : str {'MAS', 'EP', 'CAS', 'Python'}
            Specifies the publishing destination for the score code to ensure
            that compatible code is generated.

        Returns
        -------
        str
            Score code

        """

        dest = dest.upper()

        # Check for names that could result in DS2 errors.
        DS2_KEYWORDS = ['input', 'output']
        for k in DS2_KEYWORDS:
            if input_table and k == input_table.lower():
                raise ValueError(
                    'Input table name `{}` is a reserved term.'.format(input_table)
                )
            if output_table and k == output_table.lower():
                raise ValueError(
                    'Output table name `{}` is a reserved term.'.format(output_table)
                )

        # Get package code
        code = tuple(self.package.code().split('\n'))

        if dest == 'EP':
            code = (
                ('data sasep.out;',)
                + code
                + (
                    '   method run();',
                    '      set SASEP.IN;',
                    '   end;',
                    '   method term();',
                    '   end;',
                    'enddata;',
                )
            )
        elif dest == 'CAS':
            thread = DS2Thread(
                self.variables,
                input_table,
                column_names=columns,
                return_message=self.return_message,
                package=self.package,
            )

            code += (
                str(thread),
                'data SASEP.out;',
                '  dcl thread {} t;'.format(thread.name),
                '  method run();',
                '    set from t;',
                '    output;',
                '  end;',
                'enddata;',
            )

        elif dest == 'PYTHON':
            # Python code return
            code = self.package._python_code

        return '\n'.join(code)
